@@ -1,5 +1,5 @@
 #
-# $Id: Layer.pm 2005 2015-01-23 06:56:13Z gomor $
+# $Id: Layer.pm 2011 2015-02-15 17:07:47Z gomor $
 #
 package Net::Write::Layer;
 use strict;
@@ -81,24 +81,120 @@ sub _setIpHdrInclConstant {
    eval "use constant NW_IP_HDRINCL => $val;";
 }
 
-#
-# ACF - This could result in NW_AF_INET6 => 0 if neither
-# Socket6::AF_INET6 nor Socket::AF_INET6 is defined.  My change
-# doesn't preserve that, but it isn't clear that was an intended or
-# useful feature.
-#
-#sub _setAfinet6Constant {
-#   require Socket6;
-#   require Socket;
-#   my $val = 0;
-#   if (defined(&Socket6::AF_INET6)) {
-#      $val = &Socket6::AF_INET6;
-#   }
-#   elsif (defined(&Socket::AF_INET6)) {
-#      $val = &Socket::AF_INET6;
-#   }
-#   eval "use constant NW_AF_INET6  => $val;";
-#}
+sub _setAfinet6Constant {
+   my $val = 10;  # Default value, in case we don't know.
+                  # This is the value from a Ubuntu 14.10 system.
+   eval {
+      require Socket;
+      Socket->import(qw(AF_INET6));
+   };
+   if (! $@) {  # AF_INET6 constant found in Socket module.
+      $val = Socket::AF_INET6();
+   }
+   else {  # No AF_INET6 in Socket module, we try with Socket6.
+      eval {
+         require Socket6;
+         Socket6->import(qw(AF_INET6));
+      };
+      if (! $@) {   # AF_INET6 constant found in Socket6 module.
+         $val = Socket6::AF_INET6();
+      }
+   }
+
+   # If constant is not found, we stick to the default value.
+   eval "use constant NW_AF_INET6  => $val;";
+}
+
+sub _setInetPtonSub {
+   no strict 'refs';
+
+   eval {
+      require Socket;
+      Socket->import(qw(AF_INET6 inet_pton));
+   };
+   if (! $@) {  # Socket supports AF_INET6 family and inet_pton.
+      *{__PACKAGE__.'::nw_inet_pton'} = \&Socket::inet_pton;
+
+      return 1;
+   }
+
+   eval {
+      require Socket6;
+      Socket6->import(qw(AF_INET6 inet_pton));
+   };
+   if (! $@) {  # Socket6 supports AF_INET6 family and inet_pton.
+      *{__PACKAGE__.'::nw_inet_pton'} = \&Socket6::inet_pton;
+
+      return 1;
+   }
+
+   die("[-] Net::Write: inet_pton: not supported by Socket nor Socket6: ".
+       "try upgrading your Perl version or Socket/Socket6 modules.\n");
+}
+
+sub _setGetaddrinfoSub {
+   no strict 'refs';
+
+   # Try to use getaddrinfo() from main Socket module.
+   eval {
+      require Socket;
+      Socket->import(qw(AF_INET AF_INET6 getaddrinfo));
+   };
+   if (! $@) {  # Socket supports AF_INET6 family and getaddrinfo.
+      *{__PACKAGE__.'::nw_getsaddr'} = sub {
+         my ($dest, $family, $protocol, $socktype) = @_;
+
+         #print STDERR "*** Socket support OK\n";
+
+         my %hints = (
+            family => $family,
+            # If we activate that, it breaks on some OS like Mac OS X
+            #protocol => $protocol,
+            #socktype => $socktype,
+         );
+         my ($err, @res) = Socket::getaddrinfo($dest, "", \%hints);
+         if ($err) {
+            return _croak("@{[(caller(0))[3]]}: getaddrinfo: $err");
+         }
+
+         if (@res > 0) {
+            my $h = $res[0];
+            return $h->{addr};
+         }
+
+         return _croak("@{[(caller(0))[3]]}: getaddrinfo: error: $!");
+      };
+
+      return 1;
+   }
+  
+   # Main Socket module does not support getaddrinfo(), we try using Socket6
+   eval {
+      require Socket6;
+      Socket6->import(qw(AF_INET AF_INET6 getaddrinfo));
+   };
+   if (! $@) {
+      *{__PACKAGE__.'::nw_getsaddr'} = sub {
+         my ($dest, $family, $protocol, $socktype) = @_;
+
+         #print STDERR "*** Fallback on Socket6 support\n";
+
+         my @res = Socket6::getaddrinfo($dest, "", $family, $socktype)
+            or return _croak("@{[(caller(0))[3]]}: getaddrinfo: $!");
+
+         if (@res >= 5) {
+            my $saddr = $res[3];
+            return $saddr;
+         }
+
+         return _croak("@{[(caller(0))[3]]}: getaddrinfo: error: $!");
+      };
+
+      return 1;
+   }
+
+   return 1;
+}
 
 BEGIN {
    my $osname = {
@@ -107,47 +203,26 @@ BEGIN {
    };
 
    {
-     no strict 'refs';
-     *{ __PACKAGE__ . "::_check" } = $osname->{$^O} || \&_checkOther;
+      no strict 'refs';
+      *{__PACKAGE__.'::_check'} = $osname->{$^O} || \&_checkOther;
    }
+
    _setIpProtoIpConstant();
    _setIpProtoIpv6Constant();
    _setIpProtoRawConstant();
    _setIpHdrInclConstant();
-   #_setAfinet6Constant();
-}
-
-use Socket qw(:DEFAULT AF_INET);
-BEGIN {
-    # imports that may or may not be in Socket.
-    my @imports = (qw(AF_INET6 getaddrinfo));
-    my @socket6_imports;
-
-    # This might be overkill, but I'm not certain that all these imports
-    # were added to Socket at the same time.
-    for my $import (@imports) {
-        eval { Socket->import($import); };
-        if ($@) {
-            push @socket6_imports, $import;
-        }
-    }
-
-    if (@socket6_imports) {
-
-        # something we want wasn't found in Socket time to try Socket6
-        eval { require Socket6 };
-        die $@ if $@;
-        Socket6->import(@socket6_imports);
-    }
+   _setAfinet6Constant();
+   _setInetPtonSub();
+   _setGetaddrinfoSub();
 }
 
 no strict 'vars';
 
+use Socket qw(SOCK_RAW);
 use IO::Socket;
 use Net::Pcap;
 
 use constant NW_AF_INET   => AF_INET();
-use constant NW_AF_INET6  => AF_INET6();
 use constant NW_AF_UNSPEC => AF_UNSPEC();
 
 use constant NW_IPPROTO_ICMPv4 => 1;
@@ -169,10 +244,15 @@ our %EXPORT_TAGS = (
       NW_IP_HDRINCL
       NW_IPPROTO_RAW
    )],
+   subs => [qw(
+      nw_inet_pton
+      nw_getsaddr
+   )],
 );
 
 our @EXPORT_OK = (
    @{$EXPORT_TAGS{constants}},
+   @{$EXPORT_TAGS{subs}},
 );
 
 sub _checkWin32 {
@@ -194,8 +274,6 @@ sub new {
       @_,
    );
 
-   _check() or return;
-
    return $self;
 }
 
@@ -209,13 +287,14 @@ sub open {
    my $self = shift;
    my ($hdrincl) = @_;
 
-   my @res = getaddrinfo($self->[$__dst], 0, $self->[$__family], SOCK_STREAM)
-      or return _croak("@{[(caller(0))[3]]}: getaddrinfo: $!");
+   _check() or return;
+
+   my $saddr = nw_getsaddr($self->[$__dst], $self->[$__family], $self->[$__protocol])
+      or return _croak("@{[(caller(0))[3]]}: nw_getsaddr: error");
 
-   my ($family, $saddr) = @res[0, 3] if @res >= 5;
    $self->[$___sockaddr] = $saddr;
 
-   socket(my $s, $family, SOCK_RAW, $self->[$__protocol])
+   socket(my $s, $self->[$__family], SOCK_RAW(), $self->[$__protocol])
       or return _croak("@{[(caller(0))[3]]}: socket: $!");
 
    my $fd = fileno($s)
@@ -322,6 +401,12 @@ Send the raw data passed as a parameter. Returns undef on failure, true otherwis
 
 Close the descriptor.
 
+=item B<nw_getsaddr>
+
+=item B<nw_inet_pton>
+
+Internal functions.
+
 =back
 
 =head1 CONSTANTS
